#!/usr/bin/perl -w
use strict;
#use diagnostics;

#(This license is from zlib of Jean-loup Gailly and Mark Adler)
#
# Copyright (C) 2002 Frederick Dean <software@fdd.com>
#
#  This software is provided 'as-is', without any express or implied
#  warranty.  In no event will the authors be held liable for any damages
#  arising from the use of this software.
#
#  Permission is granted to anyone to use this software for any purpose,
#  including commercial applications, and to alter it and redistribute it
#  freely, subject to the following restrictions:
#
#  1. The origin of this software must not be misrepresented; you must not
#     claim that you wrote the original software. If you use this software
#     in a product, an acknowledgment in the product documentation would be
#     appreciated but is not required.
#  2. Altered source versions must be plainly marked as such, and must not be
#     misrepresented as being the original software.
#  3. This notice may not be removed or altered from any source distribution.
#
# Contributor(s):    Frederick Dean <software@fdd.com>

package session;
require db;
use Digest::MD5 qw(md5_base64);

use useradmin;

my $maxIdleSessionTime = 30 * 24 * 60 * 60;  # 30 days measured in seconds
my $maxIdleLogoutSessionTime = 24 * 60 * 60;  # one day measured in seconds

# This is a really ugly global.
# It has all the group memberships of the last validated user.
# The key is a interger group_id.
our %group = (0 => 1);  # group 0 is everyone
our $user_id;

# useful for debugging.  Prints %group
sub PrintGroupHash {
   print("<hr><font size=+1>Group hash</font><br>\n<font color=brown>\n");
   for my $key (sort(keys(%group))) {
      print("$key => $group{$key}<br>\n");
   }
   print("</font>\n");
}

# This function receives the string from the dirent groups field.
# It updates the global varaible %session::group which we use for permissions.
# The groups string is a packed list of network-byte-order shorts.
sub updateGroupGlobal {
   %group = (0 => 1);
   foreach my $id (unpack("n*",$_[0])) {
      $group{$id} = 1;
   };
}

#  ACL is a string of packed network-byte-order shorts.
#  Each rule is two shorts, a who and a flags.
#  The top 14 flag bits are permissions, and the bottom two are:
#   0x0002     1=user 0=group
#   0x0001     1=allow 0=deny
#  See access_html.pm for a description of the other flag bits.
#  User zero is special that it refers to the owner.
#  Both the owner privs and the users privs are searched in parallel, so order matters.
#  We quit on the first matching rule with the verdict of that matching rule.
sub isPrivilegedTo($$$)
{
   my($acl,$priv,$owner) = (@_);

   foreach my $rule (unpack("N*","$config::aclPrefix$acl$config::aclSuffix")) {
      my $flags = $rule & 0xffff;
      next unless($flags & $priv);  # This rule must apply to the privilege
      my $who = $rule >> 16;
      if($flags & 2) {  # if user rule
         if($who == 0) {  # if who is branch owner
            next unless defined $owner && defined $user_id && $owner == $user_id;
         } else {  # else actual user
            next unless defined $user_id && $who == $user_id;
         }
      } else {  # else group rule
         next unless $group{$who};
      };	
      return $flags & 0x1;
   }
   return 1;  # default allow (change default behaviour by editing config.pm)
}

# The web server tells us the remote IP of our browser.
# This returns the remote IP as a single integer.
sub remoteip
{
   my $remoteip = 0;
   if($ENV{'REMOTE_ADDR'} && $ENV{'REMOTE_ADDR'} =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/) {
      $remoteip = (((($1*256)+$2)*256)+$3)*256+$4;
   };
   return $remoteip;
}

# Print out the login form and exit.
sub login_html
{
   if(! $cgi::cookie{'dcvs_session'}) {
      cookie_for_login_form();
   };
   common_html::content_type("text/html");
   $cgi::keyword{'prev_username'} = $cgi::form{'username'} || "";
   $cgi::keyword{'new_username'} = $cgi::form{'new_username'} || "";
   $cgi::keyword{'new_realname'} = $cgi::form{'new_realname'} || "";
   $cgi::keyword{'new_email'} = $cgi::form{'new_email'} || "";
   $cgi::form{'username'} = undef;
   $cgi::form{'password'} = undef;

   ##### FIXME  convert form parameter to hidden elements
   
   $cgi::keyword{'warning'} ||= "";
   $cgi::keyword{"prevurl"} = $ENV{REQUEST_URI};
   print cgi::expand_template_keywords("template/login-template.html");
   exit;
}

# This will prompt the user with a login page if they are not already logged in.
# If the user does the login page correctly, it will resubmit the original HTTP GET request.
# This function will exit (not return) if a login page is displayed.
sub require_login
{
   if(!$cgi::keyword{'username'}) {
      login_html();  # show the login form
   }
}

sub require_privilege 
{
   my($acl,$priv,$owner) = (@_);

   return if(&isPrivilegedTo);  # inherit @_
   if(!$cgi::keyword{'username'}) {
      login_html();  # show the login form
      return if(&isPrivilegedTo);  # inherit @_
   }
   error_html::fatal("Sorry, you don't have privileges for this.");
}

# Call this if you require the user to belong to a group such as administrators (1).
# This will prompt for login if they are not logged in.
# If the user does not belong to the group then an error page is displayed and the routine does not return.
# The routine does not return if an error or login page was displayed.
sub require_group
{
   my($grop) = @_;

   require_login();
   return 1 if $group{$grop}; 
   db::SendSQL("SELECT name FROM groups WHERE group_id = $grop");
   my $name = db::FetchOneColumn();
   $name = $grop if ! defined $name;
   error_html::fatal("You ($cgi::keyword{'username'}) do not belong to the correct group ($name) to do this.");
   exit;
}

#  NOTE isPasswordValid() does not update $session::user_id or %session::group
sub isPasswordValid
{
   my($username,$password) = @_;

   db::SendSQL("SELECT cryptpassword, userid, disabledtext, email, realname, groups, userid ".
               "FROM user WHERE username = ".db::SqlQuote($username));
   my($cryptpassword,$userid,$disabledtext,$email,$realname,$groups) = db::FetchSQLData();
   if(!defined($cryptpassword)) {  # if invalid username
      return undef unless defined $config::minUseridToInheritFromOS;
      my($name,undef,$uid,$gid,undef,$comment,$gcos,$dir,$shell,$expire) = getpwnam($username);
      return undef unless $uid >= $config::minUseridToInheritFromOS;
      db::SendSQL("INSERT user SET cryptpassword = 'os', username = ".db::SqlQuote($username).
                  " , email = ".db::SqlQuote("$username$config::emailDomainNameForInherited").
                  ", realname = ".db::SqlQuote($gcos));
      $cryptpassword = "os";
      $email = "$username$config::emailDomainNameForInherited";
      $realname = $gcos;
      $groups = $disabledtext = "";
      db::SendSQL("SELECT LAST_INSERT_ID()");
      ($userid) = db::FetchSQLData();
   }
   my $hash;
   my $pwtype = substr($cryptpassword,0,3);
   if($pwtype eq "md5") {
      # The cryptpassword has three bytes of scheme, 22 bytes of salt, then 22 bytes of crypt password.
      my $salt = substr($cryptpassword,0,25);
      # We don't use crypt() because we want more than 8 chars of relevance in our passwords!
      $hash = md5_base64($salt . $password);
      if(substr($cryptpassword,25) ne $hash) {  # if invalid password
         return undef;
      }
   } elsif($pwtype eq "os") {  # else receive from operating system
      my $path = $INC[0];
      open(PW,"|$path/pwverify.pl >/dev/null") or exit 0;
      print(PW "$username\n$password\n");
      close(PW);
      if($? == (33<<8)) { 
         $disabledtext = "Internal Error:  I could not verify your username/password ".
                         "because of a configuration issue with pwverify.pl <br>";
      } elsif($? != (183<<8)) {
         return undef;
      }
   } else {  # else we don't know the hash used
      return undef;  # because we don't understand the crypt password cypher
   } 
   return($userid,$disabledtext,$email,$realname,$groups,$pwtype);
}

# This routine uses $cgi::form{'username'} and $cgi::form{'password'} and $cgi::cookie{'dcvs_session'}
# This routine is called when the form data contains a username or a password.
# We will abandon the old session, if any.
# Unlike isPasswordValid() this one updates the global variables  $session::user_id and %session::group
sub validate_login
{
   db::getDbNowTime();
   my $session_cookie = $cgi::cookie{'dcvs_session'};
   my $remoteip = remoteip();
   if($session_cookie) { # if they have an old cooke
      db::SendSQL("UPDATE session SET isvalid = 0, lastused = $db::now WHERE remoteip = $remoteip ".
                  "AND cookie = ". db::SqlQuote($cgi::cookie{'dcvs_session'}));
      # crumble old cookie
      #db::SendSQL("DELETE session WHERE cookie = ". db::SqlQuote($cgi::cookie{'session_cookie'}));
   } else { # else they don't allow cookies
      make_a_session_cookie();
      $cgi::cookielessFlag = 1;  # need to know when we make links
      common_html::url_keywords();  
   }
   my $username = $cgi::form{'username'};
   if(!$username) {
      $cgi::keyword{'warning'} .= "<font color=red>You must enter a username.</font><br>";
      login_html();
   }
   my $password = $cgi::form{'password'};
   if(!$password) {
      $cgi::keyword{'warning'} .= "<font color=red>You must enter a password.</font><br>";
      login_html();
   }
   my($userid,$disabledtext,$email,$realname,$groups) = isPasswordValid($username,$password);
   if(!defined($userid)) { 
      $cgi::keyword{'warning'} .= "<font color=red>Your username or password was invalid</font><br>";
      login_html();
   }
   if($disabledtext) { 
      $cgi::keyword{'warning'} .= "<font color=red>$disabledtext</font>";
      login_html;
   }
   $cgi::keyword{'username'} = $username;
   $cgi::keyword{'realname'} = $realname;
   $cgi::keyword{'email'} = $email;
   updateGroupGlobal($groups);
   $user_id = $userid;  # save in global for session::isPrivilegedTo()
   db::SendSQL("REPLACE session SET isvalid = 1, userid = $userid, remoteip = $remoteip, lastused = $db::now, ".
               "cookie = ". db::SqlQuote($cgi::cookie{'dcvs_session'}));
}

# Like validate_login(), with is_cookie_tasty() we update the globals $session::user_id and %session::group.
sub is_cookie_tasty
{
   my $session_cookie = $cgi::cookie{'dcvs_session'};
   return if !defined($session_cookie);   # if no cookie, then we don't know the taste.
   my $remoteip = remoteip();
   db::getDbNowTime();
   db::SendSQL("SELECT lastused, isvalid, username, realname, disabledtext, email, groups ".
               "FROM session LEFT JOIN user ON session.userid = user.userid ".
               "WHERE session.cookie = ".db::SqlQuote($session_cookie) . " ".
               "AND session.remoteip = $remoteip ");
   my($lastused,$isvalid,$username,$realname,$disabledtext,$email,$groups) = db::FetchSQLData();
   # If the cookie didn't match the database, change it because it may have come from a stale bookmark, or worse a web page.
   if(!defined($lastused) && $cgi::cookielessFlag) {
      cookie_for_login_form();
      return;
   };
   if(!defined($lastused) || $lastused < ($db::now - $maxIdleSessionTime)) {  # if session has timed out
      # Clean out idle sessions.
      # We could do this more frequently if we wanted to, but this seems frequent enough.
      db::SendSQL("DELETE FROM session WHERE lastused < " . ($db::now - $maxIdleSessionTime)); 
      db::SendSQL("DELETE FROM session WHERE isvalid = 0 AND lastused < " . ($db::now - $maxIdleLogoutSessionTime)); 
      return;
   }
   if(! $isvalid) {
      return;  # and it is still invalid
   };
   if($disabledtext) {
      #$cgi::keyword{'warning'} .= "<font color=red>$disabledtext</font>";
      #db::SendSQL("REPLACE session SET isvalid = 0, userid = $userid, remoteip = $remoteip, lastused = $db::now, ".
      #            "cookie = ". db::SqlQuote($cgi::cookie{'dcvs_session'}));
      return;
   }
   # Define these things so the code knows we are logged in.
   $cgi::keyword{'username'} = $username;
   $cgi::keyword{'realname'} = $realname;
   $cgi::keyword{'email'} = $email;
   db::SendSQL("SELECT userid FROM user WHERE username = ".db::SqlQuote($username));
   $user_id = db::FetchOneColumn();  # save in global for session::isPrivilegedTo()
   # Update our session idle time.
   db::SendSQL("UPDATE session SET session.lastused = $db::now WHERE cookie = ". db::SqlQuote($session_cookie) );
   updateGroupGlobal($groups);
}

sub make_a_session_cookie
{
   my $data = time() . $$ . rand() . \@_;
   my $hash = md5_base64($data);
   $hash = md5_base64($hash . rand());
   $hash =~ tr/+\//_-/;  # make chars url friendly 
   $cgi::cookie{'dcvs_session'} = $hash;
   my $remoteip = remoteip();
   db::getDbNowTime();
   db::SendSQL("REPLACE session SET cookie = ".db::SqlQuote($hash).", remoteip = $remoteip, ".
               "lastused = $db::now, userid = 0");
}

sub cookie_for_login_form
{
   make_a_session_cookie();
   my $cookiepath = $ENV{'SCRIPT_NAME'} || "/";
   print "Set-Cookie: dcvs_session=$cgi::cookie{'dcvs_session'} ; path=$cookiepath; expires=Sun, 30-Jun-2029 00:00:00 GMT\n";
   # When they come back with the cookie, we will ignore the URI cookie, but until then assume the will not honor the cookie cookie.
   $cgi::cookilessFlag = 1;
   common_html::url_keywords();
}

# After calling this, old pages will also not work (i.e. use of back button).
# ...except of course resubmitting the login form.
sub logout
{
   if($cgi::cookie{'dcvs_session'}) {
      db::SendSQL("UPDATE session SET isvalid = 0 WHERE remoteip = " . db::SqlQuote(remoteip()) .
                  " AND cookie = " . db::SqlQuote($cgi::cookie{'dcvs_session'}));
   };
   $cgi::keyword{'username'} = $cgi::keyword{'realname'} = $cgi::keyword{'email'} = "";
   $user_id = undef;
   %group = ( 0 => 1 );  # only group "everyone"
}

# This is the big external function.
# Calling this will set up user information if the user is already logged in.
sub check
{
   $cgi::keyword{'loginlogout'} = cgi::Link("","loginform=1") . "Login</a>";
   $cgi::keyword{'username'} = $cgi::keyword{'realname'} = $cgi::keyword{'email'} = "";
   if($cgi::form{'loginform'}) {  # if GET is requesting the login form
      login_html();  # show the login form.
   }
   if(defined($cgi::form{'username'}) || defined($cgi::form{'password'})) {  # if user just submitted the login form
      validate_login();
   } elsif($cgi::form{'logout'}) {
      logout();
   } else {
      is_cookie_tasty()
   }
   if(defined($cgi::form{'login'}) && $cgi::form{'login'} =~ /Create/i) {
      error_html::fatal("Sorry, we don't support email created accounts yet.");
      exit;
   }
   if($cgi::keyword{'username'}) {
      $cgi::keyword{'loginlogout'} = cgi::Link("","logout=1") . "Logout</a> as $cgi::keyword{'username'}";
   }
}

1;

__END__

We keep a session with a 32byte cookie.  This cookie is either
sent as a cookie, or as a bogus first directory.  We know if
they need the bogus directory because we try to set a cookie
when we give the login form.  If the login form parameters
come back without a cookie we use the directory.

For the rest of the program, they can tell if we are logged in
because $cgi::keyword{'usernmae'} $cgi::keyword{'realname'} are defined.

We need to keep track of all the cookie we give out so we can
tell if one has timed out.  When it does we better change
it because it may have been virtual-directory-encoded and
send in an email.








